/*
 * deviced
 *
 * Copyright (c) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <vconf.h>
#include <libgdbus/dbus-system.h>

#include "core/log.h"
#include "core/device-notifier.h"
#include "core/udev.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "extcon/extcon.h"

#define METHOD_GET_CRADLE	"GetCradle"
#define SIGNAL_CRADLE_STATE	"ChangedCradle"

static struct extcon_ops cradle_extcon_ops;

static void cradle_send_broadcast(int status)
{
	static int old;
	int ret;

	if (old == status)
		return;

	_I("broadcast cradle status %d", status);
	old = status;

	ret = dbus_handle_broadcast_dbus_signal_var(DEVICED_PATH_SYSNOTI,
						DEVICED_INTERFACE_SYSNOTI,
						SIGNAL_CRADLE_STATE,
						g_variant_new("(i)", status));
	if (ret < 0)
		_E("Failed to send dbus signal(%s)", SIGNAL_CRADLE_STATE);
}

static int cradle_update(int status)
{
	_I("jack - cradle changed %d", status);
	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(INTERNAL_LOCK_CRADLE, LCD_NORMAL);
	cradle_send_broadcast(status);
	if (vconf_set_int(VCONFKEY_SYSMAN_CRADLE_STATUS, status) != 0) {
		_E("failed to set vconf status");
		return -EIO;
	}

	if (status == DOCK_SOUND) {
		if (disp_plgn.pm_lock_internal)
			disp_plgn.pm_lock_internal(INTERNAL_LOCK_CRADLE, LCD_DIM, STAY_CUR_STATE, 0);
	} else if (status == DOCK_NONE) {
		if (disp_plgn.pm_unlock_internal)
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_CRADLE, LCD_DIM, PM_SLEEP_MARGIN);
	}

	return 0;
}

static int display_changed(void *data)
{
	enum state_t state;
	int cradle;

	if (!data)
		return 0;

	state = *(int *)data;
	if (state != S_NORMAL)
		return 0;

	cradle = cradle_extcon_ops.status;
	if (cradle == DOCK_SOUND) {
		if (disp_plgn.pm_lock_internal)
			disp_plgn.pm_lock_internal(INTERNAL_LOCK_CRADLE, LCD_DIM, STAY_CUR_STATE, 0);
		_I("sound dock is connected! dim lock is on.");
	}

	return 0;
}

GVariant *dbus_cradle_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	return g_variant_new("(i)", cradle_extcon_ops.status);
}

static const dbus_method_s dbus_methods[] = {
	{ METHOD_GET_CRADLE,     NULL, "i", dbus_cradle_handler },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_SYSNOTI,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static void cradle_init(void *data)
{
	int ret;

	register_notifier(DEVICE_NOTIFIER_LCD, display_changed);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_SYSNOTI, &dbus_interface);
	if (ret < 0)
		_E("fail to init dbus method(%d)", ret);
}

static void cradle_exit(void *data)
{
	unregister_notifier(DEVICE_NOTIFIER_LCD, display_changed);
}

static struct extcon_ops cradle_extcon_ops = {
	.name   = EXTCON_CABLE_DOCK,
	.init   = cradle_init,
	.exit   = cradle_exit,
	.update = cradle_update,
};

EXTCON_OPS_REGISTER(cradle_extcon_ops)
